@@ -1,16 +1,13 @@
-extern crate llvm_tools;
-extern crate target_build_utils;
 extern crate walkdir;
 
+use std::borrow::Cow;
 use std::env;
 use std::ffi::OsStr;
 use std::ffi::OsString;
 use std::fs::File;
 use std::io::Write;
 use std::path::{Path, PathBuf};
-use std::process;
 use std::process::Command;
-use target_build_utils::TargetInfo;
 use walkdir::{DirEntry, WalkDir};
 
 fn build_hermit(src_dir: &Path, target_dir_opt: Option<&Path>) {
@@ -18,18 +15,28 @@ fn build_hermit(src_dir: &Path, target_dir_opt: Option<&Path>) {
 		src_dir.exists(),
 		"rusty_hermit source folder does not exist"
 	);
-	let target = TargetInfo::new().expect("Could not get target info");
+	let target_arch = env::var_os("CARGO_CFG_TARGET_ARCH").unwrap();
 	let profile = env::var("PROFILE").expect("PROFILE was not set");
 	let mut cmd = Command::new("cargo");
 
-	if target.target_arch() == "x86_64" {
+	cmd.env_remove("RUSTUP_TOOLCHAIN");
+	if option_env!("RUSTC_WORKSPACE_WRAPPER")
+		.unwrap_or_default()
+		.ends_with("clippy-driver")
+	{
+		cmd.env("RUSTC_WORKSPACE_WRAPPER", "clippy-driver");
+	}
+
+	cmd.env("CARGO_TERM_COLOR", "always");
+
+	if target_arch == "x86_64" {
 		cmd.current_dir(src_dir)
 			.arg("build")
 			.arg("-Z")
 			.arg("build-std=core,alloc")
 			.arg("--target")
 			.arg("x86_64-unknown-none-hermitkernel");
-	} else if target.target_arch() == "aarch64" {
+	} else if target_arch == "aarch64" {
 		cmd.current_dir(src_dir)
 			.arg("build")
 			.arg("-Z")
@@ -55,6 +62,11 @@ fn build_hermit(src_dir: &Path, target_dir_opt: Option<&Path>) {
 	// disable all default features
 	cmd.arg("--no-default-features");
 
+	if target_arch == "aarch64" {
+		cmd.arg("--features");
+		cmd.arg("aarch64-qemu-stdout");
+	}
+
 	// do we have to enable PCI support?
 	#[cfg(feature = "pci")]
 	{
@@ -91,41 +103,40 @@ fn build_hermit(src_dir: &Path, target_dir_opt: Option<&Path>) {
 	}
 
 	let mut rustflags = vec!["-Zmutable-noalias=no".to_string()];
+	let outer_rustflags = env::var("CARGO_ENCODED_RUSTFLAGS").unwrap();
 
 	#[cfg(feature = "instrument")]
 	{
 		rustflags.push("-Zinstrument-mcount".to_string());
-		// Add outer `RUSTFLAGS` to command
-		if let Ok(var) = env::var("RUSTFLAGS") {
-			rustflags.push(var);
-		}
+		// Add outer rustflags to command
+		rustflags.push(outer_rustflags);
 	}
 
 	#[cfg(not(feature = "instrument"))]
 	{
 		// If the `instrument` feature feature is not enabled,
-		// filter it from outer `RUSTFLAGS` before adding them to the command.
-		if let Ok(var) = env::var("RUSTFLAGS") {
-			let flags = var
-				.split(',')
+		// filter it from outer rustflags before adding them to the command.
+		if !outer_rustflags.is_empty() {
+			let flags = outer_rustflags
+				.split('\x1f')
 				.filter(|&flag| !flag.contains("instrument-mcount"))
 				.map(String::from);
 			rustflags.extend(flags);
 		}
 	}
 
-	cmd.env("RUSTFLAGS", rustflags.join(","));
+	cmd.env("CARGO_ENCODED_RUSTFLAGS", rustflags.join("\x1f"));
 
 	let status = cmd.status().expect("failed to start kernel build");
 	assert!(status.success());
 
-	let lib_location = if target.target_arch() == "x86_64" {
+	let lib_location = if target_arch == "x86_64" {
 		target_dir
 			.join("x86_64-unknown-none-hermitkernel")
 			.join(&profile)
 			.canonicalize()
 			.unwrap() // Must exist after building
-	} else if target.target_arch() == "aarch64" {
+	} else if target_arch == "aarch64" {
 		target_dir
 			.join("aarch64-unknown-hermit")
 			.join(&profile)
@@ -138,59 +149,38 @@ fn build_hermit(src_dir: &Path, target_dir_opt: Option<&Path>) {
 
 	let lib = lib_location.join("libhermit.a");
 
-	rename_symbol("rust_begin_unwind", &lib);
-	rename_symbol("rust_oom", &lib);
+	let mut symbols = vec!["rust_begin_unwind", "rust_oom"];
 
-	#[cfg(feature = "mem")]
-	{
-		for symbol in ["memcpy", "memmove", "memset", "memcmp", "bcmp"] {
-			rename_symbol(symbol, &lib);
-		}
+	if target_arch == "aarch64" {
+		symbols.extend(include_str!("aarch64-duplicate-symbols").lines());
 	}
 
+	rename_symbols(symbols, &lib);
+
 	println!("cargo:rustc-link-search=native={}", lib_location.display());
 	println!("cargo:rustc-link-lib=static=hermit");
 
-	//HERMIT_LOG_LEVEL_FILTER sets the log level filter at compile time
-	// Doesn't actually rebuild atm - see: https://github.com/rust-lang/cargo/issues/8306
+	// HERMIT_LOG_LEVEL_FILTER sets the log level filter at compile time
 	println!("cargo:rerun-if-env-changed=HERMIT_LOG_LEVEL_FILTER");
 }
 
 /// Kernel and user space has its own versions of panic handler, oom handler, memcpy, memset, etc,
 /// Consequently, we rename the functions in the libos to avoid collisions.
-/// In addition, it provides us the offer to create a optimized version of memcpy
-/// in user space.
-fn rename_symbol(symbol: impl AsRef<OsStr>, lib: impl AsRef<Path>) {
-	// Get access to llvm tools shipped in the llvm-tools-preview rustup component
-	let llvm_tools = match llvm_tools::LlvmTools::new() {
-		Ok(tools) => tools,
-		Err(llvm_tools::Error::NotFound) => {
-			eprintln!("Error: llvm-tools not found");
-			eprintln!("Maybe the rustup component `llvm-tools-preview` is missing?");
-			eprintln!("  Install it through: `rustup component add llvm-tools-preview`");
-			process::exit(1);
-		}
-		Err(err) => {
-			eprintln!("Failed to retrieve llvm-tools component: {:?}", err);
-			process::exit(1);
-		}
-	};
+fn rename_symbols(symbols: impl IntoIterator<Item = impl AsRef<OsStr>>, lib: impl AsRef<Path>) {
+	let args = symbols.into_iter().flat_map(|symbol| {
+		let option = OsStr::new("--redefine-sym");
+		let arg = [symbol.as_ref(), "=kernel-".as_ref(), symbol.as_ref()]
+			.into_iter()
+			.collect::<OsString>();
+		[Cow::Borrowed(option), Cow::Owned(arg)]
+	});
 
-	// Retrieve path of llvm-objcopy
-	let llvm_objcopy = llvm_tools
-		.tool(&llvm_tools::exe("llvm-objcopy"))
-		.expect("llvm-objcopy not found in llvm-tools");
-
-	// Rename symbols
-	let arg = IntoIterator::into_iter([symbol.as_ref(), "=kernel-".as_ref(), symbol.as_ref()])
-		.collect::<OsString>();
-	let status = Command::new(llvm_objcopy)
-		.arg("--redefine-sym")
-		.arg(arg)
+	let status = Command::new("rust-objcopy")
+		.args(args)
 		.arg(lib.as_ref())
 		.status()
-		.expect("failed to execute llvm-objcopy");
-	assert!(status.success(), "llvm-objcopy was not sucessful");
+		.expect("Failed to execute rust-objcopy. Is cargo-binutils installed?");
+	assert!(status.success(), "rust-objcopy was not successful");
 }
 
 #[cfg(all(not(feature = "rustc-dep-of-std"), not(feature = "with_submodule")))]
@@ -202,7 +192,7 @@ fn build() {
 	let src_dir = out_dir.join("rusty-hermit");
 
 	if !src_dir.as_path().exists() {
-		let status = Command::new("cargo")
+		let status = Command::new(env!("CARGO"))
 			.current_dir(out_dir)
 			.arg("download")
 			.arg("--output")
@@ -251,7 +241,7 @@ fn configure_cargo_rerun_if_changed(src_dir: &Path) {
 
 	WalkDir::new(src_dir)
 		.into_iter()
-		.filter_entry(|e| is_not_ignored(e))
+		.filter_entry(is_not_ignored)
 		.filter_map(|v| v.ok())
 		.filter_map(|v| v.path().canonicalize().ok())
 		.for_each(|x| println!("cargo:rerun-if-changed={}", x.display()));
